/*
Very simple GLUT program using bare OpenGL libraries to
demonstrate use of GL_ARB_shading_language_100.

Orion Sky Lawlor, olawlor@acm.org, 2006/09/15 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

/* Variables updated by the GUI */
mat4 modelMat(1.0), viewMat(1.0), animMat(1.0);
mat4 animMatSum(1.0);

float distance=1.0, blurry=1.7;
float lens_size=0.05, focal_plane=2.0;

float spin_x, spin_y, zoom, mandelfunk1, mandelfunk2;

int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}

void Exit(const char *where,const char *why) {
	fprintf (stderr, "FATAL OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

/************* Object we'll be drawing ***********/
class mainObject_t {
public:
	mainObject_t();
	void draw(void);
};


/** Set up a 3D model, stored in a vertex buffer */
mainObject_t::mainObject_t()
{
	/* nothing to set up here... */
}

/** A camera capable of doing antialiased rendering */
class aaCamera {
public:
	/* Camera's center of projection
	   FIXME: what about isometric cameras? */
	vec3 center;
	aaCamera(const vec3 &center_) :center(center_) {}

	/* Draw an antialiased triangle with these vertices */
	void tri(const vec3 &V1,const vec3 &V2,const vec3 &V3) const {
		vec3 V[3]; V[0]=V1; V[1]=V2; V[2]=V3; tri(V);
	}
	/* Draw an antialiased triangle with these three vertices */
	void tri(const vec3 *V) const;
};

/* Draw an antialiased triangle with these three vertices */
void aaCamera::tri(const vec3 *V) const
{
	enum {n=3}; /* number of vertices */
	vec3 PN=cross(V[1]-V[0],V[2]-V[0]);// plane normal
	glNormal3fv(PN);
	for (int i=0;i<n;i++) {
		/* Pass the entire triangle in for each vertex.
		   If only we had a "triangle shader", we wouldn't have to 
		   duplicate so much work in each vertex shader. */
		for (int j=0;j<n;j++) {
			glMultiTexCoord4fv(GL_TEXTURE3+j,vec4(V[j],1.0));
		}
		/* Compute unit-length, outward-facing edges of triangle */
		vec3 EL=V[i]-V[(i+1)%n];
		vec3 ER=V[i]-V[(i+2)%n];
		glMultiTexCoord4fv(GL_TEXTURE6,vec4(EL,0.0));
		glMultiTexCoord4fv(GL_TEXTURE7,vec4(ER,0.0));
		glVertex3fv(V[i]);
	}
}


/** Main drawing routine, called by main_display below */
void mainObject_t::draw(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	float fov_deg=100.0;
	gluPerspective(fov_deg, win_w/(float)win_h,0.01,100.0);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0,0,-2); /* push stuff down the X axis */
	glLoadMatrixf(&modelMat[0][0]);
	glMultMatrixf(&animMatSum[0][0]);
	
	// Each frame, fade in a little of the animMat...
	animMatSum=animMatSum+0.01*(animMatSum*animMat-animMatSum);
	
	oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	p->begin();
	
	/* radians across the entire field of view */
	float fov_rads=2.0*tan(fov_deg*0.5*3.14159265/180.0);
	/* size of blurred region, in radians */
	float blursize=fov_rads/win_h*blurry;
	p->set("blursize_antialias",blursize);
	p->set("focal_plane_inv",1.0/focal_plane); /* distance from camera where everything's in focus */
	p->set("lens_size",lens_size); /* diameter of camera lens, world units */
	vec3 cameraloc=vec3(transpose(viewMat)*vec4(0,0,+distance,1));
	p->set("cameraloc",cameraloc);
	p->set("cameradir",-normalize(cameraloc));
	aaCamera cam(cameraloc);
	
	
	glBegin(GL_TRIANGLES);
	glColor3f(0,0,1); /* blue */
	float dx,dy;
#define  LOOP_XY(ddx,ddy) \
	for (dy=-6.0;dy<=6.0;dy+=ddy) \
	for (dx=-6.0;dx<=6.0;dx+=ddx) 
	
	LOOP_XY(1,2)
		cam.tri(vec3(dx+0,dy+0,-2), vec3(dx+1,dy+0,-2), vec3(dx+0,dy+2,-2));
	
	glColor3f(0,1,0); /* green */
	LOOP_XY(2,1)
		cam.tri(vec3(dx+0,dy+0,-1), vec3(dx+2,dy+0,-1), vec3(dx+0,dy+1,-1));
	
	glColor3f(1,0,0); /* red */
	LOOP_XY(1,1)
		cam.tri(vec3(dx+0,dy+0,0), vec3(dx+1,dy+0,0), vec3(dx+0,dy+1,0));
	
	glEnd();
	
	/* Stop using programmable shader */
	p->end();
	delete p;
	
	/* This line causes the world to repeatedly redraw itself, using 100% of the CPU. */
	glutPostRedisplay();
}


/******** Program Initialization *********/
mainObject_t *mainObject=0; /**< Big object we're trying to draw */
/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	Check("Entering display routine");
	// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Set up a little error-prone OpenGL state:
	glEnable(GL_BLEND);
	           /*  src color  + dest color  */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_DEPTH_TEST); //<- z-buffer fighting stinks.
	
	// Draw stuff
	Check("About to draw");
	mainObject->draw();
	Check("Done drawing");
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h-50;
	glViewport(0, 0, (GLsizei)w, (GLsizei)(h)); //<- hack: leave room at top for GLUI spinner boxes
}

int main(int argc, char** argv)
{
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	glutInitWindowSize(800,600);
	int main_window = glutCreateWindow("GLSL Mandelbrot Demo");
	
	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();

	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	// Make an object to draw
	mainObject=new mainObject_t;
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_TOP);
	/* This causes the main window to be redrawn when a GLUI control changes */
	glui->set_main_gfx_window( main_window );
	(new GLUI_Spinner( glui, "Distance:", &distance))
		->set_float_limits(0.0,10.0);
	(new GLUI_Spinner( glui, "Antialiasing:", &blurry))
		->set_float_limits(0.001, 100.0);
	(new GLUI_Spinner( glui, "Focal Plane:", &focal_plane))
		->set_float_limits(0.001, 10.0);
	(new GLUI_Spinner( glui, "Lens Size:", &lens_size))
		->set_float_limits(0.0001, 1.0);
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"Animate:",&animMat[0][0]);
	
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
